from collections.abc import Iterable

import hdf5plugin  # noqa: F401
import numpy as np
from ewokscore import Task
from ewoksfluo.io import hdf5
from ewoksfluo.io import output_uri
from ewoksfluo.tasks import xrf_results
from silx.io import h5py_utils
from silx.io import url

# from ewokscore import Task


# from ewoksfluo.io.hdf5 import ReadHdf5File
# from ewoksfluo.tasks import xrf_results

DEFAULTS = {"output_root_group": "weighted"}


class WeightedSumResults(
    Task,
    input_names=[
        "bliss_scan_uris",
        "output_root_uris",
        "xrf_results_uris",
        "output_root_groups",
        "output_root_uri",
    ],
    optional_input_names=["output_root_group"],
    output_names=[
        "xrf_results_uri",
        "bliss_scan_uri",
        "output_root_uri",
        "output_root_group",
    ],
):
    """Add single-scan XRF results of multiple detectors"""

    def _read_xrf_results(self, uri):
        params, massfr = {}, {}

        uri = url.DataUrl(uri)

        with h5py_utils.open_item(uri.file_path(), uri.data_path()) as results:

            for group in results["massfractions"]:
                params[group] = np.array(results["parameters"][group])
                massfr[group] = np.array(results["massfractions"][group])

        return massfr, params

    def _dict_op(self, fun, a, b):
        ka = set(a.keys())
        kb = set(b.keys())

        s = dict()

        for k in set.intersection(ka, kb):
            s[k] = fun(a[k], b[k])

        return s

    def dict_op(self, fun, *op):
        if len(op) <= 1:
            return op

        res = self._dict_op(fun, op[0], op[1])

        for i in range(2, len(op)):
            res = self._dict_op(fun, res, op[i])

        return res

    def dict_plus(self, *op):
        return self.dict_op(lambda a, b: a + b, *op)

    def dict_mult(self, *op):
        return self.dict_op(lambda a, b: a * b, *op)

    def run(self) -> None:
        params = {**DEFAULTS, **self.get_input_values()}

        xrf_results_uris = params["xrf_results_uris"]
        output_root_uri = params["output_root_uri"]
        output_root_group = params["output_root_group"]
        summed_fit = None
        summed_prod = None

        for xrf_results_uri in xrf_results_uris:
            massfraction, parameters = self._read_xrf_results(xrf_results_uri)

            if summed_fit is None:
                summed_fit = parameters
            else:
                summed_fit = self.dict_plus(summed_fit, parameters)

            if summed_prod is None:
                summed_prod = self.dict_mult(massfraction, parameters)
            else:
                summed_prod = self.dict_plus(
                    summed_prod, self.dict_mult(massfraction, parameters)
                )

        weighted = self._dict_op(lambda a, b: a / b, summed_prod, summed_fit)

        for group in weighted:
            weighted[group][np.isnan(weighted[group])] = 0

        config = {
            "xrf_results_uris": xrf_results_uris,
            "bliss_scan_uris": params["bliss_scan_uris"],
        }

        _, scan_h5path = hdf5.split_h5uri(output_root_uri)
        output_root_uri = output_uri.compose_full_output_uri(
            self.inputs.output_root_uri,
            default_output_data_path=scan_h5path,
            extra_data_paths=(self.inputs.output_root_group, "weighted"),
        )

        xrf_results.save_xrf_results(
            output_root_uri,
            config,
            summed_fit,
            None,
            weighted,
        )

        self.outputs.bliss_scan_uri = self.inputs.output_root_uri
        self.outputs.output_root_uri = self.inputs.output_root_uri
        self.outputs.output_root_group = output_root_group
        self.outputs.xrf_results_uri = output_root_uri + "/results"


class UrisAggregator(
    Task,
    optional_input_names=[
        "bliss_scan_uri_0",
        "output_root_uri_0",
        "output_root_group_0",
        "detector_name_0",
        "xrf_results_uri_0",
        "bliss_scan_uri_1",
        "output_root_uri_1",
        "output_root_group_1",
        "detector_name_1",
        "xrf_results_uri_1",
    ],
    output_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "output_root_group",
        "detector_name",
        "xrf_results_uri",
    ],
):

    def run(self):

        names = (
            "bliss_scan_uri",
            "output_root_uri",
            "detector_name",
            "xrf_results_uri",
            "output_root_group",
        )

        for n in names:
            output = []

            for i in range(2):
                val = self.get_input_value(f"{n}_{i}", [])

                if isinstance(val, str):
                    output += [
                        val,
                    ]
                elif isinstance(val, Iterable):
                    output += [*val]
                else:
                    raise RuntimeError(f"{val} must be string or iterable of strings")

            print(n, output)

            setattr(self.outputs, f"{n}", output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 10 20:40:04 2026

@author: blissadm
"""

import os
import time

from ewokscore import Task
from PyMca5.PyMcaIO import ConfigDict

DEFAULTS = {
    "batchconfig_ext": ".batch.cfg",
    "batch_force": {
        "fit.stripflag": 0,
        "fit.fitweight": 0,
        "fit.linearfitflag": 1,
    },
    "retryPeriod": 3,
    "retryN": 1200,
    "force_overwrite": False,
    "output_root_group": None,
}


class PatchConfigFile(
    Task,
    input_names=[
        "bliss_scan_uri",
        "detector_name",
        "config_file",
        "output_root_uri",
    ],
    optional_input_names=[
        "batchconfig_ext",
        "batch_force",
        "retryPeriod",
        "retryN",
        "force_overwrite",
        "output_root_group",
    ],
    output_names=[
        "bliss_scan_uri",
        "detector_name",
        "output_root_uri",
        "config_file",
        "output_root_group",
    ],
):
    def run(self):

        pars = {**DEFAULTS, **self.get_input_values()}

        self.outputs.bliss_scan_uri = pars["bliss_scan_uri"]
        self.outputs.output_root_uri = pars["output_root_uri"]
        self.outputs.detector_name = pars["detector_name"]
        self.outputs.output_root_group = pars["output_root_group"]
        config_file = pars["config_file"]
        batch_ext = pars["batchconfig_ext"]
        batch_force = pars["batch_force"]
        force_overwrite = pars["force_overwrite"]

        retryPeriod = pars["retryPeriod"]
        retryN = pars["retryN"]

        for i in range(retryN):
            if os.path.isfile(config_file):
                break
            time.sleep(retryPeriod)
        else:
            raise RuntimeError(f"Config file {config_file} not found!")

        root, ext = os.path.splitext(config_file)

        batch_savefile = root + batch_ext

        if force_overwrite or not os.path.isfile(batch_savefile):
            cfg = ConfigDict.ConfigDict(filelist=config_file)

            for k in batch_force:
                _c = cfg
                kk = k.split(".")
                for _k in kk[:-1]:
                    _c = _c[_k]

                _c[kk[-1]] = batch_force[k]

            cfg.write(batch_savefile)

        self.outputs.config_file = batch_savefile



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os

import numpy as np
from esrf_pathlib import ESRFPath
from ewokscore import Task
from silx.io import h5py_utils


class ExtractRanddisp(
    Task,
    input_names=[
        "bliss_scan_uri",
    ],
    optional_input_names=[
        "output_filename",
        "subscans",
    ],
    output_names=["bliss_scan_uri", "output_filename"],
):
    """
    This task extract the random displacement to text.
    """

    def run(self):
        bliss_uri = self.inputs["bliss_scan_uri"]
        uri = bliss_uri.split("::/")

        bliss_filename = uri[0]
        scan_id = uri[1]

        scan_no = int(scan_id.split(".")[0])

        esrf_path = ESRFPath(bliss_filename)

        output_filename = os.path.join(
            esrf_path.processed_dataset_path,
            "projections",
            f"{esrf_path.collection}_{esrf_path.dataset}_{scan_no:04d}.txt",
        )

        output_filename = self.get_input_value("output_filename", output_filename)

        output_path = os.path.abspath(os.path.dirname(output_filename))

        if not os.path.isdir(output_path):
            os.makedirs(output_path, exist_ok=True)
            os.chmod(
                output_path, 0o770
            )  # nosec B103: group write required for shared pipeline

        types = []

        # Wait for scan completion
        with h5py_utils.open_item(
            bliss_filename, f"{scan_id}/end_time", retry_invalid=True
        ):
            pass

        print("OK!")

        print(bliss_filename, scan_id)

        subs = self.get_input_value("subscans", None)

        if subs is None:
            # open the tomo seauence, read the configuration and scan sequence
            with h5py_utils.open_item(bliss_filename, scan_id) as tomo_sequence:
                scan_numbers = tomo_sequence["subscans"]["scan_numbers"][:]

                for k in tomo_sequence["technique"]["subscans"]:
                    typ = tomo_sequence["technique"]["subscans"][k]["type"][()]
                    if isinstance(typ, bytes):
                        typ = typ.decode()
                    types += [
                        typ,
                    ]
        else:
            scan_numbers = list(subs.keys())
            types = list(subs.values())

        print(types)

        disp_y_px = []
        disp_z_px = []

        for i, n in enumerate(scan_numbers):
            if types[i] == "tomo:step":  # Only valid for projections

                with h5py_utils.open_item(bliss_filename, f"/{n}.1") as scan:
                    disp_y_px += [scan["technique/proj/rand_pos_y_px"][()]]
                    disp_z_px += [scan["technique/proj/rand_pos_z_px"][()]]

            elif types[i] == "tomo:return_ref":  # add 3 zeros for return ref

                disp_y_px += [
                    np.zeros((3,), dtype=np.float32),
                ]
                disp_z_px += [
                    np.zeros((3,), dtype=np.float32),
                ]

        disp_y_px = np.concatenate(disp_y_px)
        disp_z_px = np.concatenate(disp_z_px)

        np.savetxt(output_filename, np.array([disp_y_px, -disp_z_px]).T, "%.03f")

        self.outputs["bliss_scan_uri"] = bliss_uri
        self.outputs["output_filename"] = output_filename



from configparser import ConfigParser

import numpy as np
from ewokscore import Task
from ewoksfluo.tasks import nexus_utils
from ewoksfluo.tasks.hdf5_utils import create_hdf5_link
from ewoksfluo.tasks.hdf5_utils import link_bliss_scan
from silx.io import h5py_utils

DEFAULTS = {
    "counter_uri_template": "instrument/%s/data",
    "output_counter_name": "photons_i0",
    "max_ref": 1e15,
    "min_ref": 1e8,
    "gamma_coef": 2.0,
    "bliss_slow_subscan": 2,
    "output_root_group": None,
}


class NormalizeCurrent(
    Task,
    input_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "fastscan_counter_name",
        "slowscan_counter_name",
        "dwelltime_counter_name",
        "diode_factor",
    ],
    optional_input_names=[
        "gamma_coef",
        "min_ref",
        "max_ref",
        "counter_uri_template",
        "output_counter_name",
        "bliss_slow_subscan",
        "output_root_group",
    ],
    output_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "output_root_group",
        "counter_name",
    ],
):
    def run(self):
        start_time = nexus_utils.now()
        params = {**DEFAULTS, **self.get_input_values()}

        bliss_scan_uri = params["bliss_scan_uri"]
        bliss_slow_subscan = params["bliss_slow_subscan"]
        output_root_uri = params["output_root_uri"]
        counter_template = params["counter_uri_template"]
        min_ref = params["min_ref"]
        max_ref = params["max_ref"]
        gamma_coef = params["gamma_coef"]
        diode_factor = params["diode_factor"] * 1e12  # From ph/pA to ph/A
        output_root_group = params["output_root_group"]

        slow_counter_name = params["slowscan_counter_name"]
        fast_counter_name = params["fastscan_counter_name"]
        dwelltime_counter_name = params["dwelltime_counter_name"]

        output_counter_name = params["output_counter_name"]

        fast_scan_uris = bliss_scan_uri.split("::")
        filename = fast_scan_uris[0]
        fast_scanno = fast_scan_uris[1]

        slow_scanno = fast_scanno.split(".")[0] + f".{bliss_slow_subscan}"

        # Read data from slow scan
        with h5py_utils.open_item(
            filename, slow_scanno
        ) as slow_scan:  # type: ignore[reportGeneralTypeIssues]
            ref_values = np.array(slow_scan[counter_template % slow_counter_name])

        # Read data from fast scan
        with h5py_utils.open_item(filename, fast_scanno) as fast_scan:  # type: ignore[reportGeneralTypeIssues]
            # epoch  = np.array(fd[ref_counter_template%self.inputs.get("data_epoch_counter", "epoch_trig")])
            values = np.array(fast_scan[counter_template % fast_counter_name])
            dwelltime = np.mean(fast_scan[counter_template % dwelltime_counter_name])

        # Filter slow scan values
        ref_msk = np.logical_and(
            ref_values * diode_factor > min_ref, ref_values * diode_factor < max_ref
        )  # When shutter was open and not saturated

        ref_values = ref_values[ref_msk]

        # Compute the conversion factor
        gamma = np.mean(values) / (np.mean(ref_values) * dwelltime)
        real_gamma_exp = np.log10(gamma / gamma_coef)
        gamma_exp = np.round(real_gamma_exp)

        gamma = gamma_coef * 10**gamma_exp

        if output_root_group is not None:
            lvls = (fast_scanno, output_root_group, output_counter_name)
        else:
            lvls = (fast_scanno, output_counter_name)

        # Save in process group
        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config={
                "real_gamma_exp": real_gamma_exp,
                "gamma_exp": gamma_exp,
                "diode_factor": diode_factor / 1e12,
                "gamma_coef": gamma_coef,
            },
            default_levels=lvls,
        ) as (process_group, already_existed):

            outentry = process_group
            for i in range(len(lvls) - 1):
                outentry = outentry.parent

            if not already_existed:
                link_bliss_scan(outentry, bliss_scan_uri)

                nxdata = nexus_utils.create_nxdata(
                    process_group, "normcounter", signal="data"
                )
                dset = nxdata.create_dataset("data", data=values / gamma * diode_factor)
                dset.attrs["interpretation"] = "spectrum"

                nxdetector = outentry["instrument"].create_group(output_counter_name)
                nxdetector.attrs["NX_class"] = "NXdetector"
                create_hdf5_link(nxdetector, "data", dset)
                create_hdf5_link(outentry["measurement"], output_counter_name, dset)

            output_root_uri = f"{outentry.file.filename}::{outentry.name}"

        self.outputs.bliss_scan_uri = output_root_uri
        self.outputs.output_root_uri = output_root_uri
        self.outputs.output_root_group = output_root_group
        self.outputs.counter_name = output_counter_name


class NormalizationFactorFromConfig(
    Task,
    input_names=["bliss_scan_uri", "output_root_uri", "config_file"],
    optional_input_names=[
        "xrf_results_uri",
        "detector_name",
        "output_root_group",
    ],
    output_names=[
        "bliss_scan_uri",
        "detector_name",
        "output_root_uri",
        "xrf_results_uri",
        "output_root_group",
        "counter_normalization_template",
        "counter_normalization_factor",
        "config_file",
    ],
):
    def get_from_cfg(self, cfg):
        cp = ConfigParser()
        cp.read(cfg)

        flux = cp.getfloat("concentrations", "flux")
        dwelltime = cp.getfloat("concentrations", "time")

        matrix = cp.get("attenuators", "Matrix")
        matdat = matrix.replace(" ", "").split(",")
        matrix_composition = matdat[1]
        matrix_density = float(matdat[2])
        matrix_thickness = float(matdat[3])
        print("INFO FROM CONFIG FILE")
        print("Config input flux: {0:g} (ph/s)".format(flux))
        print("Config dwell time: {0} (s)".format(dwelltime))
        print("Matrix composition: {0}".format(matrix_composition))
        print("Matrix density: {0} (g/cm**3)".format(matrix_density))
        print("Matrix thickness: {0} (cm)".format(matrix_thickness))
        areal_dens_ratio = matrix_density * matrix_thickness * 1e7  # ng/mm**2
        return flux, areal_dens_ratio, dwelltime

    def run(self):
        self.outputs.bliss_scan_uri = self.inputs.bliss_scan_uri
        self.outputs.xrf_results_uri = self.get_input_value("xrf_results_uri", None)
        self.outputs.output_root_uri = self.inputs.output_root_uri
        self.outputs.detector_name = self.get_input_value("detector_name", None)
        self.outputs.config_file = self.inputs.config_file
        self.outputs.output_root_group = self.get_input_value("output_root_group", None)

        cfg_flux, areal_dens_ratio, cfg_dwelltime = self.get_from_cfg(
            self.inputs.config_file
        )

        factor = cfg_flux * cfg_dwelltime * areal_dens_ratio

        self.outputs.counter_normalization_factor = factor

        self.outputs.counter_normalization_template = (
            f"{factor:.04e}/<instrument/{{}}/data>"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  9 16:12:33 2026

@author: blissadm
"""

from silx.io import h5py_utils


def wait_scan_finished(bliss_scan_uri):

    scanuris = bliss_scan_uri.split("::")

    filename = scanuris[0]
    scanno = scanuris[1]

    with h5py_utils.open_item(
        filename,
        f"/{scanno}/end_time",
        retry_invalid=True,
        retry_timeout=3600,
        retry_period=5,
    ):
        pass



import os

import h5py
from ewokscore import Task
from silx.io import h5py_utils
from silx.io.url import DataUrl

DEFAULTS = {"resolve_links": True}


class AddEntryToMasterfile(
    Task,
    input_names=["masterfile_uri", "output_root_uri"],
    optional_input_names=["resolve_path"],
    output_names=["masterfile_uri", "output_root_uri"],
):
    """
    This task add or replace an entry in the masterfile
    """

    def _resolve_h5_link(self, output_root_uri, already_visited=None):
        if already_visited is None:
            already_visited = []

        output_root_uri_txt = (
            output_root_uri.file_path() + "::" + output_root_uri.data_path()
        )

        if output_root_uri_txt in already_visited:
            raise RuntimeError(
                f"{output_root_uri} already visited, must have a circular linking issue!"
            )

        already_visited += [
            output_root_uri_txt,
        ]

        with h5py_utils.open_item(output_root_uri.file_path(), "/") as fd:
            obj = fd.get(output_root_uri.data_path(), getlink=True)

            if isinstance(obj, h5py.SoftLink):
                linkval = obj.link

                if not os.path.isabs(linkval):
                    linkval = obj.path.normpath(
                        os.path.join(output_root_uri.data_path(), linkval)
                    )

                return self._resolve_h5_link(
                    DataUrl(output_root_uri.file_path() + "::" + linkval),
                    already_visited,
                )

            elif isinstance(obj, h5py.ExternalLink):
                return self._resolve_h5_link(
                    DataUrl(obj.path + "::" + obj.link), already_visited
                )

        return output_root_uri

    def run(self):

        pars = {**DEFAULTS, **self.get_input_values()}

        muri = pars["masterfile_uri"].split("::")
        ouri = pars["output_root_uri"].split("::")

        masterfile_uri = DataUrl(file_path=muri[0], data_path=muri[1])
        output_root_uri = DataUrl(file_path=ouri[0], data_path=ouri[1])

        if pars["resolve_links"]:
            output_root_uri = self._resolve_h5_link(output_root_uri)

        relpath = os.path.relpath(
            output_root_uri.file_path(), os.path.dirname(masterfile_uri.file_path())
        )

        with h5py.File(masterfile_uri.file_path(), "a") as fd:
            if masterfile_uri.data_path() in fd:
                del fd[masterfile_uri.data_path()]

            fd[masterfile_uri.data_path()] = h5py.ExternalLink(
                relpath, output_root_uri.data_path()
            )

        self.outputs["masterfile_uri"] = self.inputs["masterfile_uri"]
        self.outputs["output_root_uri"] = self.inputs["output_root_uri"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul  6 20:10:52 2025

@author: blissadm
"""

import os

from esrf_pathlib import ESRFPath
from ewokscore import Task
from pyicat_plus.client.main import IcatClient
from silx.io.url import DataUrl


class PublishProcessed(
    Task,
    input_names=["output_root_uri", "icat_url"],
    optional_input_names=["bliss_scan_uri", "metadata"],
    output_names=["output_root_uri"],
):
    """Publish processed data to data portal"""

    def run(self):

        output_root_uri = DataUrl(self.inputs.output_root_uri)
        bliss_scan_uri = self.get_input_value("bliss_scan_uri", None)

        processed_data_folder = os.path.dirname(output_root_uri.file_path())
        processed_data = ESRFPath(processed_data_folder)

        if bliss_scan_uri is None:
            raw_data = processed_data.raw_dataset_file
        else:
            bliss_scan_uri = DataUrl(bliss_scan_uri)
            raw_data = ESRFPath(bliss_scan_uri.file_path())

        metadata = self.get_input_value("metadata", {})

        if "Sample_name" not in metadata:
            metadata["Sample_name"] = raw_data.collection

        client = IcatClient(metadata_urls=self.inputs["icat_url"])
        client.store_processed_data(
            beamline=raw_data.beamline,
            proposal=raw_data.proposal,
            dataset=f"{raw_data.collection}_{raw_data.dataset}",
            path=processed_data_folder,
            metadata=metadata,
            raw=raw_data.raw_dataset_path,
        )

        self.outputs.output_root_uri = (
            output_root_uri.file_path() + "::" + (output_root_uri.data_path() or "/")
        )

    def _fix_data_path(self, path):
        if path.startswith("/mnt/multipath-shares/data"):
            path = path.replace("/mnt/multipath-shares/data", "/data")

        return path



import os

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pint
from ewokscore import Task

# from PIL import Image
from silx.io import h5py_utils
from silx.io import url
from tifffile import RESUNIT
from tifffile import imwrite

ureg = pint.UnitRegistry()

DEFAULTS = {
    "do_gallery": False,
    "output_template": "IMG_{elm}_areal_density_ng_mm2_{scanno:04d}",
}


class TiffExporterFromRegrid(
    Task,
    input_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "xrf_results_uri",
        "output_root_group",
    ],
    optional_input_names=[
        "do_gallery",
        "output_template",
    ],
    output_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "xrf_results_uri",
        "output_root_group",
    ],
):
    def run(self):

        params = {**DEFAULTS, **self.get_input_values()}

        self.outputs.bliss_scan_uri = params["bliss_scan_uri"]
        self.outputs.output_root_uri = params["output_root_uri"]
        self.outputs.xrf_results_uri = params["xrf_results_uri"]
        self.outputs.output_root_group = params["output_root_group"]

        do_gallery = params["do_gallery"]
        output_template = params["output_template"]
        output_root_group = params["output_root_group"]

        xrf_results_uri = url.DataUrl(params["xrf_results_uri"])

        output_path = os.path.dirname(xrf_results_uri.file_path())
        output_gallery_path = os.path.join(output_path, "gallery")

        os.makedirs(output_path, exist_ok=True)
        os.chmod(
            output_path, 0o770
        )  # nosec B103: group write required for shared pipeline

        if do_gallery:
            os.makedirs(output_gallery_path, exist_ok=True)
            os.chmod(
                output_gallery_path, 0o770
            )  # nosec B103: group write required for shared pipeline

        bliss_scan_uri = url.DataUrl(self.inputs.bliss_scan_uri)

        uri = bliss_scan_uri.data_path()
        uri = uri.split("/")[1]
        scanno = int(uri.split(".")[0])

        print(xrf_results_uri)

        with h5py_utils.open_item(xrf_results_uri.file_path(), xrf_results_uri.data_path()) as _res:  # type: ignore[reportGeneralTypeIssues]

            quanti = False

            if "massfractions" in _res:
                _res = _res["massfractions"]
                quanti = True
            elif "parameters" in _res:
                _res = _res["parameters"]
            else:
                raise RuntimeError(
                    "Found none of massfractions or parameters to be exported!"
                )

            grps = []
            if "signal" in _res.attrs:
                grps += [
                    _res.attrs["signal"],
                ]

            if "auxiliary_signals" in _res.attrs:
                grps += list(_res.attrs["auxiliary_signals"])

            kwargs = {}
            axes = []
            axnames = []
            if "axes" in _res.attrs:
                resolutions = []
                resolutions_units = []

                for a in _res.attrs["axes"]:
                    axnames += [
                        _res[a].attrs["long_name"],
                    ]
                    d = np.array(_res[a])
                    axes += [
                        d,
                    ]

                    res = float(np.mean(np.diff(d)))
                    resu = _res[a].attrs["units"]

                    resolutions_units += [
                        resu,
                    ]

                    ps = ureg.Quantity(res, resu)

                    if ps.check("[length]"):
                        resolutions += [
                            1.0 / ps.to("cm").magnitude,
                        ]
                    else:
                        resolutions += [
                            1e4 / ps.magnitude,
                        ]

                kwargs["resolutionunit"] = RESUNIT.CENTIMETER
                kwargs["resolution"] = tuple(resolutions)
                axes = axes[::-1]
                axnames = axnames[::-1]

            cmap = "jet"

            for g in grps:

                data = np.array(_res[g], dtype=np.float32)
                #                im = Image.fromarray(data, mode="F")
                gg = g.replace(" ", "_")
                templatedata = {
                    "elm": gg,
                    "detector_name": output_root_group,
                    "scanno": scanno,
                }
                #                im.save(
                #                    os.path.join(
                #                        output_path, output_template.format(**templatedata) + ".tiff"
                #                    ),
                #                    **kwargs,
                #                )

                imwrite(
                    os.path.join(
                        output_path, output_template.format(**templatedata) + ".tiff"
                    ),
                    data,
                    **kwargs,
                )

                if output_gallery_path is not None:
                    norm = mpl.colors.Normalize(0, np.nanmax(data))
                    f, ax = plt.subplots()
                    ax.pcolor(*axes, data, norm=norm, cmap=cmap, shading="nearest")
                    ax.invert_yaxis()

                    if "deg" not in resolutions_units:
                        ax.set_aspect("equal", "box")

                    if len(axnames) == 2:
                        ax.set_xlabel(axnames[0])
                        ax.set_ylabel(axnames[1])

                    f.suptitle(g)
                    f.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                        ax=ax,
                        label=(
                            "Areal mass density ($ng/mm^2$)" if quanti else "Parameters"
                        ),
                    )
                    f.savefig(
                        os.path.join(
                            output_gallery_path,
                            output_template.format(**templatedata) + ".jpg",
                        )
                    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov  1 17:02:10 2025

@author: blissadm
"""

import os

import imageio
import numpy as np
from ewokscore import Task
from silx.io import h5py_utils

DEFAULTS = {"detector_name": None, "target_width": 200, "max_size": 200 * 1024}


class AnimatedWebP(
    Task,
    input_names=[
        "nx_file",
    ],
    optional_input_names=["detector_name"],
    output_names=["nx_file", "detector_name"],
):
    """Export an animated WebP file from a nx-fluotomo"""

    def bin_stack(self, img, bin_y=2, bin_x=2):
        n, h, w = img.shape

        h2 = h // bin_y * bin_y
        w2 = w // bin_x * bin_x
        img = img[:, :h2, :w2]

        return img.reshape(n, h2 // bin_y, bin_y, w2 // bin_x, bin_x).mean(axis=(2, 4))

    def run(self):

        pars = {
            **DEFAULTS,
            **self.get_input_values(),
        }

        nx_file = pars["nx_file"]
        detector_name = pars["detector_name"]
        t_width = pars["target_width"]
        max_size = pars["max_size"]

        processed_data_folder = os.path.dirname(nx_file)
        galleryfolder = os.path.join(processed_data_folder, "gallery")
        os.makedirs(galleryfolder, exist_ok=True)
        os.chmod(
            galleryfolder, 0o770
        )  # nosec B103: group write required for shared pipeline

        with h5py_utils.open_item(nx_file, "/") as fd:
            for e in fd:

                if detector_name is not None and not e.startswith(detector_name):
                    continue

                data = fd[e]["data/data"][()]
                data[np.isnan(data)] = 0

                width = data.shape[2]

                binning = int(np.round(width / t_width))

                if binning < 0:
                    binning = 1

                data = self.bin_stack(data, binning, binning)

                mn = np.min(data)
                Mx = np.max(data)

                converted = np.array((data - mn) / (Mx - mn) * 255, dtype=np.uint8)

                #                imageio.v3.imwrite(
                #                    os.path.join(galleryfolder, f"{e}.webp"), converted,  quality=40, method=4,
                #                )

                encoded = imageio.v3.imwrite(
                    "<bytes>",
                    converted,
                    fps=5,
                    lossless=False,
                    extension=".webp",
                    quality=70,
                    method=4,
                )

                size = len(encoded)

                if size < max_size:
                    with open(os.path.join(galleryfolder, f"{e}.webp"), "wb") as f:
                        f.write(encoded)

        self.outputs.nx_file = nx_file
        self.outputs.detector_name = detector_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 25 16:19:55 2026

@author: blissadm
"""

import os
from datetime import datetime as dt

import numpy as np
import pint
from esrf_pathlib import ESRFPath
from ewokscore import Task
from nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey
from silx.io import h5py_utils

ureg = pint.UnitRegistry()


def _tostr(val):
    if isinstance(val, str):
        return val
    elif isinstance(val, bytes):
        return val.decode()
    else:
        return str(val)


DEFAULTS = {
    "include_positioners": True,
    "scan_list": None,
    "motors_mapping": {
        "x_translation": ["isy", "spy"],
        "y_translation": ["isz", "spz"],
        "z_translation": [
            "sx",
        ],
        "rotation_angle": [
            "somega",
        ],
    },
    "motors_units": {
        "isy": "nm",
        "isz": "nm",
    },
    "wait_scan": None,
}


class Fluo2Nx(
    Task,
    input_names=[
        "masterfile_path",
        "output_file",
        "detector_name",
    ],
    optional_input_names=[
        "include_positioners",
        "scan_list",
        "motors_mapping",
        "motors_units",
        "wait_scan",
    ],
    output_names=["output_file", "detector_name"],
):
    """
    This task takes a fulltomo sequence and creates the associated NXtomo file.
    """

    def run(self):

        pars = {**DEFAULTS, **self.get_input_values()}

        masterfile_path = pars["masterfile_path"]
        scan_list = pars["scan_list"]
        output_file = pars["output_file"]
        motors_mapping = pars["motors_mapping"]
        detector_name = pars["detector_name"]
        include_positioners = pars["include_positioners"]
        motors_units = pars["motors_units"]
        wait_scan = pars["wait_scan"]

        output_path = os.path.abspath(os.path.dirname(output_file))
        esrf_path = ESRFPath(masterfile_path)

        if not os.path.isdir(output_path):
            os.makedirs(output_path, exist_ok=True)
            os.chmod(
                output_path, 0o770
            )  # nosec B103: group write required for shared pipeline

        shape = None
        egy = None

        if wait_scan is not None:
            with h5py_utils.open_item(
                masterfile_path,
                f"{wait_scan}.1_fluofit",
                retry_invalid=True,
                retry_timeout=3600,
            ) as fd:
                pass

        with h5py_utils.open_item(masterfile_path, "/") as fd:

            # Look for the most represented set of shapes
            shapes = {}
            for k in fd:

                if not k.endswith("_fluofit"):
                    continue

                scan_no = int(k.split(".")[0])

                if scan_list is not None and scan_no not in scan_list:
                    continue

                scan = fd[k]
                regrid = scan[f"{detector_name}/regrid/results/massfractions"]

                for elm in regrid:
                    shp = regrid[elm].shape

                    if len(shp) == 2:
                        break

                if shp not in shapes:
                    shapes[shp] = {"N": 0, "scans": []}

                shapes[shp]["N"] += 1
                shapes[shp]["scans"] += [
                    scan_no,
                ]

            nmax = 0
            for k, v in shapes.items():
                if v["N"] > nmax:
                    shape = k
                    scan_list = v["scans"]

            # At this point, we have a consistent list of scans with the same shape.
            scan_list = sorted(scan_list)

            # Get positioners
            positioners = {}
            positioners_units = {}
            rays = set()

            fast_pixel_size = None
            slow_pixel_size = None

            fast_motor_name = None
            slow_motor_name = None
            count_time = []

            for i, scan_no in enumerate(scan_list):
                scan = fd[f"{scan_no}.1_fluofit"]

                if egy is None:
                    if (
                        "metadata" in scan["instrument"]
                        and "InstrumentMonochromator_energy"
                        in scan["instrument/metadata"]
                    ):
                        egy = ureg.Quantity(
                            float(
                                _tostr(
                                    scan[
                                        "instrument/metadata/InstrumentMonochromator_energy"
                                    ][()]
                                )
                            ),
                            "keV",
                        )
                    elif "InstrumentMonochromator" in scan["instrument"]:
                        egy = ureg.Quantity(
                            float(
                                _tostr(
                                    scan["instrument/InstrumentMonochromator/energy"][
                                        ()
                                    ]
                                )
                            ),
                            "keV",
                        )

                for k in scan["instrument/positioners_start"]:
                    ds = scan["instrument/positioners_start"][k]

                    if k not in positioners:
                        positioners[k] = (
                            np.empty((len(scan_list),), dtype=np.float64) * np.nan
                        )
                        positioners_units[k] = motors_units.get(
                            k, ds.attrs.get("units", None)
                        )

                    positioners[k][i] = ds[()]

                regrid = scan[f"{detector_name}/regrid/results/massfractions"]

                for elm in regrid:
                    if regrid[elm].shape == shape:
                        rays.add(elm)

                count_time += [
                    float(scan["instrument/fscan_parameters/acq_time"][()].decode()),
                ]

                if fast_pixel_size is None:
                    fast_pixel_size = float(
                        scan["instrument/fscan_parameters/fast_step_size"][()].decode()
                    )

                if slow_pixel_size is None:
                    slow_pixel_size = float(
                        scan["instrument/fscan_parameters/slow_step_size"][()].decode()
                    )

                if fast_motor_name is None:
                    fast_motor_name = scan["instrument/fscan_parameters/fast_motor"][
                        ()
                    ].decode()

                if slow_motor_name is None:
                    slow_motor_name = scan["instrument/fscan_parameters/slow_motor"][
                        ()
                    ].decode()

            if fast_motor_name == "spzp":
                x_pixel_size = ureg.Quantity(
                    slow_pixel_size, positioners_units[slow_motor_name]
                )
                y_pixel_size = ureg.Quantity(
                    fast_pixel_size, positioners_units[fast_motor_name]
                )
            else:
                x_pixel_size = ureg.Quantity(
                    fast_pixel_size, positioners_units[fast_motor_name]
                )
                y_pixel_size = ureg.Quantity(
                    slow_pixel_size, positioners_units[slow_motor_name]
                )

            # Start populating the NXtomo object
            nxtomo = NXtomo()
            nxtomo.start_time = dt.now()
            nxtomo.bliss_original_files = tuple(
                f"{masterfile_path}::{s}.1_fluofit" for s in scan_list
            )

            nxtomo.energy = egy
            # TODO
            nxtomo.instrument.detector.tomo_n = len(scan_list)
            #            nxtomo.instrument.detector.distance =
            nxtomo.instrument.detector.x_pixel_size = x_pixel_size
            nxtomo.instrument.detector.y_pixel_size = y_pixel_size
            nxtomo.sample.x_pixel_size = x_pixel_size
            nxtomo.sample.y_pixel_size = y_pixel_size
            #            nxtomo.instrument.source.distance =
            #            nxtomo.sample.propagation_distance =

            nxtomo.sample.name = esrf_path.collection
            nxtomo.title = esrf_path.dataset

            #            nxtomo.control.data =

            nxtomo.instrument.name = esrf_path.beamline
            nxtomo.instrument.detector.count_time = pint.Quantity(
                np.array(count_time), "s"
            )
            nxtomo.instrument.detector.image_key_control = [
                ImageKey.PROJECTION,
            ] * len(scan_list)

            for k, mots in motors_mapping.items():

                values = pint.Quantity(
                    np.zeros((len(scan_list),), dtype=np.float64),
                    "degree" if k == "rotation_angle" else "mm",
                )

                for m in mots:
                    #                    print(m, positioners_units[m])
                    values += pint.Quantity(positioners[m], positioners_units[m])

                setattr(nxtomo.sample, k, values)

            # Read data and save
            data = np.empty((len(scan_list), *shape), dtype=np.float32)
            for elm in rays:

                data *= np.nan

                for i, scan_no in enumerate(scan_list):
                    scan = fd[f"{scan_no}.1_fluofit"]
                    regrid = scan[f"{detector_name}/regrid/results/massfractions"]

                    if elm in regrid:
                        data[i] = regrid[elm][()]

                    data[np.isnan(data)] = 0

                nxtomo.instrument.detector.data = (
                    data * 1e-7
                )  # Conversion from ng/mm2 to g/cm2
                nxtomo.end_time = dt.now()
                nxtomo.save(output_file, f"{detector_name}_{elm}", overwrite=True)

                print(f"{detector_name}_{elm} saved")

        if include_positioners:
            for elm in rays:
                with h5py_utils.open_item(
                    output_file, f"{detector_name}_{elm}", mode="a"
                ) as fd:

                    grp = fd.require_group("instrument/positioners")
                    grp.attrs["NX_class"] = "NXcollection"

                    for k, v in positioners.items():
                        if k in grp:
                            del grp[k]

                        grp[k] = v
                        if positioners_units[k] is not None:
                            grp[k].attrs["units"] = positioners_units[k]

        self.outputs.output_file = output_file
        self.outputs.detector_name = detector_name



# -*- coding: utf-8 -*-


import hdf5plugin  # noqa: F401
import numpy as np
from ewokscore import Task
from ewoksfluo.tasks import nexus_utils
from silx.io import h5py_utils
from silx.io import url

DEFAULTS = {
    "output_root_group": None,
    "fscan_parameters": "instrument/fscan_parameters",
    "instrument_positioners_template": "instrument/positioners/{}",
}


class FscanRegrid(
    Task,
    input_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "xrf_results_uri",
    ],
    optional_input_names=[
        "output_root_group",
        "fscan_parameters",
        "instrument_positioners_template",
    ],
    output_names=[
        "xrf_results_uri",
        "bliss_scan_uri",
        "output_root_uri",
        "output_root_group",
    ],
):
    """Regrid XRF results from fscan parameters (No taking care of units of slow motor)."""

    def _get_motor_unit(self, scan, motor, instrument_data_template):

        if instrument_data_template.format(f"axis_{motor}") in scan:
            ds = scan[instrument_data_template.format(f"axis_{motor}")]

            #            print(ds.attrs)
            return ds.attrs.get("units", "")
        elif instrument_data_template.format(motor) in scan:
            ds = scan[instrument_data_template.format(motor)]
            #            print(ds.attrs)
            return ds.attrs.get("units", "")

        return ""

    def run(self):

        start_time = nexus_utils.now()
        params = {
            **DEFAULTS,
            **self.get_input_values(),
        }

        bliss_scan_uri = url.DataUrl(params["bliss_scan_uri"])
        output_root_uri = url.DataUrl(params["output_root_uri"])
        xrf_results_uri = url.DataUrl(params["xrf_results_uri"])
        output_root_group = params["output_root_group"]
        fscan_parameters = params["fscan_parameters"]
        instrument_positioners_template = params["instrument_positioners_template"]

        with h5py_utils.open_item(bliss_scan_uri.file_path(), bliss_scan_uri.data_path()) as scan:  # type: ignore[reportGeneralTypeIssues]

            fp = scan[fscan_parameters]

            fastmot = fp["fast_motor"][()].decode()
            slowmot = fp["slow_motor"][()].decode()
            mode = fp["fast_motor_mode"][()].decode()
            fast_n = int(fp["fast_npoints"][()].decode())
            slow_n = int(fp["slow_npoints"][()].decode())
            fast_start = float(fp["fast_start_pos"][()].decode())
            slow_start = float(fp["slow_start_pos"][()].decode())
            fast_step_size = float(fp["fast_step_size"][()].decode())
            slow_step_size = float(fp["slow_step_size"][()].decode())

            fastmotu = self._get_motor_unit(
                scan, fastmot, instrument_positioners_template
            )
            slowmotu = self._get_motor_unit(
                scan, slowmot, instrument_positioners_template
            )

        fast_motpos = fast_start + (np.arange(fast_n) + 0.5) * fast_step_size
        slow_motpos = slow_start + (np.arange(slow_n)) * slow_step_size

        # Load data
        datas = {}
        shape = None
        dtype = None
        dats = {"massfractions", "parameters", "uncertainties"}

        with h5py_utils.open_item(xrf_results_uri.file_path(), xrf_results_uri.data_path()) as res:  # type: ignore[reportGeneralTypeIssues]

            for dd in dats:
                if dd in res:
                    datas[dd] = {}
                    for k in res[dd]:
                        datas[dd][k] = res[dd][k][()]

                        if shape is None:
                            shape = datas[dd][k].shape

                        if dtype is None:
                            dtype = datas[dd][k].dtype

        A = np.arange(fast_n * slow_n)
        #        A.shape = (fast_n, slow_n)
        A.shape = (slow_n, fast_n)

        if mode == "ZIGZAG":
            A[1::2, :] = A[1::2, :][:, ::-1]

        lvls = (
            (
                output_root_uri.data_path(),
                output_root_group,
                "regrid",
            )
            if output_root_group is not None
            else (
                output_root_uri.data_path(),
                "regrid",
            )
        )

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config={
                "mode": mode,
                "fast_motor": fastmot,
                "slow_motor": slowmot,
            },
            default_levels=lvls,
        ) as (process_group, already_existed):

            if already_existed:
                del process_group["results"]

            results = process_group.require_group("results")
            results.attrs["NX_class"] = "NXcollection"

            for dd in datas:

                groups = [k for k in datas[dd].keys()]

                nxdata = nexus_utils.create_nxdata(
                    results,
                    dd,
                    signal=groups[0],
                )
                nxdata.attrs["auxiliary_signals"] = groups[1:]
                nxdata.attrs["axes"] = [slowmot, fastmot]
                nxdata.attrs["interpretation"] = "image"

                dsfm = nxdata.create_dataset(fastmot, data=fast_motpos)
                dssm = nxdata.create_dataset(slowmot, data=slow_motpos)

                dsfm.attrs["long_name"] = f"{fastmot} ({fastmotu})"
                dsfm.attrs["units"] = fastmotu

                dssm.attrs["long_name"] = f"{slowmot} ({slowmotu})"
                dssm.attrs["units"] = slowmotu

                nxdata.create_dataset(f"{fastmot}_indices", data=np.int64(1))
                nxdata.create_dataset(f"{slowmot}_indices", data=np.int64(0))

                for k in datas[dd]:
                    dset = nxdata.create_dataset(k, shape=A.shape, dtype=dtype)
                    dset[:] = datas[dd][k][A]

        self.outputs.bliss_scan_uri = params["bliss_scan_uri"]
        self.outputs.xrf_results_uri = (
            output_root_uri.file_path() + "::" + ("/".join(lvls)) + "/results"
        )
        self.outputs.output_root_uri = params["output_root_uri"]
        self.outputs.output_root_group = params["output_root_group"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 26 17:04:36 2026

@author: blissadm
"""

import os

import h5py
from ewokscore import Task
from pyicat_plus.metadata.drac_report_builder import DracReportBuilder
from silx.io import h5py_utils
from silx.io.url import DataUrl

DEFAULTS = {}


class SumfitSummaryReport(
    Task,
    input_names=["output_root_uri", "bliss_scan_uri"],
    optional_input_names=[],
    output_names=["output_root_uri", "bliss_scan_uri"],
):
    """Generate a report of sumfit tasks for the data portal"""

    def run(self):

        pars = {**DEFAULTS, **self.get_input_values()}

        # From https://pyicat-plus.readthedocs.io/en/stable/other_features.html#drac-report-builder

        output_root_uri = DataUrl(pars["output_root_uri"])
        gallery_path = os.path.join(
            os.path.dirname(output_root_uri.file_path()), "gallery"
        )

        fitting_results = {}
        detectors = set()

        with h5py_utils.open_item(
            output_root_uri.file_path(), output_root_uri.data_path()
        ) as scan:
            for k in scan:

                if isinstance(scan[k], h5py.Group) and "sumfit" in scan[k]:
                    detectors.add(k)
                    fitgrp = scan[k]["sumfit"]
                    for elm in fitgrp["areal_density"]:

                        if elm not in fitting_results:
                            fitting_results[elm] = {}

                        fitting_results[elm][k] = (
                            fitgrp["areal_density"][elm][()],
                            fitgrp["fitarea"][elm][()],
                        )

        table_results = []
        table_title = ["Element"]

        for d in detectors:
            table_title += [
                f"{d} fit area",
                f"{d} areal density (ng/mm2)",
            ]

        if len(detectors) > 1:
            table_title += ["Total fit area", "Average areal density (ng/mmg2)"]

        for elm, v in fitting_results.items():
            area = 0.0
            prod = 0.0

            row = [
                elm,
            ]

            for d in detectors:
                vals = v.get(d, None)

                if vals is None:
                    row += ["N/A", "N/A"]
                else:
                    area += vals[1]
                    prod += vals[0] * vals[1]

                    row += [f"{vals[1]:.5g}", f"{vals[0]:.5g}"]

            if len(detectors) > 1:
                weighted = prod / area
                row += [f"{area:.5g}", f"{weighted:.5g}"]

            table_results += [
                row,
            ]

        # Create report
        report = DracReportBuilder(
            dataset_type="xrf_fitting",
            dataset_title="XRF fitting report",
        )

        # Add text elements
        #        report.add_info("This is a demo info item.")
        #        report.add_warning("Warning !")

        # Add a row of images (must be inside an image list)
        report.start_image_list()

        for d in detectors:
            filepath = os.path.join(gallery_path, f"{d}_sumfit.jpg")
            if os.path.isfile(filepath):

                report.add_image(
                    path_to_image=filepath,
                    image_title=f"{d} sumfit",
                    thumbnail_height=200,
                )

        report.end_image_list()

        # Add a table
        report.add_table(
            title="Fitting results",
            columns=table_title,
            data=table_results,
            orientation="vertical",
        )

        # Export DRAC JSON
        report.render_json(
            path_to_json_dir=gallery_path,
        )

        self.outputs.output_root_uri = pars["output_root_uri"]
        self.outputs.bliss_scan_uri = pars["bliss_scan_uri"]



import os
import time

# import time
from threading import Lock

import matplotlib.pyplot as plt
import numpy as np
from ewokscore import Task

# from ewoksfluo.io.hdf5 import split_h5uri
from ewoksfluo.tasks import nexus_utils
from PyMca5.PyMcaIO import ConfigDict
from PyMca5.PyMcaPhysics.xrf.ClassMcaTheory import ClassMcaTheory
from PyMca5.PyMcaPhysics.xrf.ConcentrationsTool import ConcentrationsTool
from silx.io import h5py_utils
from silx.io import url

# from ewoksfluo.tasks.hdf5_utils import create_hdf5_link
# from ewoksfluo.tasks.hdf5_utils import link_bliss_scan
# from esrf_pathlib import ESRFPath

_mcaLock = Lock()

DEFAULTS = {
    "gallery_filename": None,
    "instrument_data_template": "instrument/{}/data",
    "config_file_timeout": 3600,
}


class AdvancedFitSumSingleDetector(
    Task,
    input_names=[
        "bliss_scan_uri",
        "detector_name",
        "config_file",
        "output_root_uri",
        "photon_counter_name",
        "output_root_group",
    ],
    optional_input_names=[
        "gallery_filename",
        "instrument_data_template",
        "config_file_timeout",
    ],
    output_names=[
        "bliss_scan_uri",
        "detector_name",
        "output_root_uri",
        "config_file",
        "output_root_group",
    ],
):
    def run(self):

        pars = {**DEFAULTS, **self.get_input_values()}

        bliss_scan_uri = url.DataUrl(pars["bliss_scan_uri"])
        output_root_uri = url.DataUrl(pars["output_root_uri"])
        detector_name = pars["detector_name"]
        config_file = pars["config_file"]
        dettmpl = pars["instrument_data_template"]
        gallery_filename = pars["gallery_filename"]
        photon_counter = pars["photon_counter_name"]
        config_timeout = pars["config_file_timeout"]
        output_root_group = pars["output_root_group"]

        # Wait for config file
        t0 = time.perf_counter()
        while not os.path.isfile(config_file):
            t1 = time.perf_counter()

            if t1 - t0 > config_timeout:
                raise TimeoutError("Timeout waiting for config file.")

            time.sleep(5)

        with h5py_utils.open_item(
            bliss_scan_uri.file_path(), bliss_scan_uri.data_path()
        ) as scan:

            data = np.array(scan[dettmpl.format(detector_name)], dtype=np.float64)
            photons = np.array(scan[dettmpl.format(photon_counter)], dtype=np.float64)

            avg_dwtime = float(scan["instrument/fscan_parameters/acq_time"][()])

        avg_photons = np.mean(photons)

        avg_flux = avg_photons / avg_dwtime

        rel_factor = photons / avg_photons
        rel_factor.shape = rel_factor.shape[0], 1

        ndata = min(rel_factor.shape[0], data.shape[0])

        rel_factor = rel_factor[:ndata]
        data = data[:ndata]

        data = np.sum(data * rel_factor, keepdims=True, axis=0) / data.shape[0]

        output_config_file = os.path.join(
            os.path.dirname(output_root_uri.file_path()), f"{detector_name}.cfg"
        )
        cfg = ConfigDict.ConfigDict(filelist=config_file)
        cfg["concentrations"]["flux"] = avg_flux
        cfg["concentrations"]["time"] = avg_dwtime
        cfg.write(output_config_file)

        matdat = cfg["attenuators"]["Matrix"]
        # print(matrix)
        # matdat = matrix.replace(" ", "").split(",")
        matrix_density = float(matdat[2])  # g/cm3
        matrix_thickness = float(matdat[3])  # cm
        matrix_factor = matrix_density * matrix_thickness * 1e7  # ng/mm**2

        with open(output_config_file, "r") as fd:
            config_txt = fd.read()

        with _mcaLock:
            mcafit = ClassMcaTheory(output_config_file)

        mcafit.setData(x=np.arange(data.shape[1]), y=data)
        mcafit.estimate()
        p, fit = mcafit.startfit(digest=1)

        # ct = ConcentrationsTool(mcafit.config, {"result": fit})
        # conc = ct.processFitResult()

        # print(conc)

        concentrationsTool = ConcentrationsTool()

        # the concentrations tool has its own configuration
        concentrationsConfig = concentrationsTool.configure()

        # however, the simplest way to use it is to use the configuration
        # set when sending the fit configuration
        concentrationsConfig = fit["config"]["concentrations"]
        # or the one used with the fit
        # concentrationsConfig = mcafit.configure()["concentrations"]

        conc, info = concentrationsTool.processFitResult(
            config=concentrationsConfig,
            fitresult={"result": fit},
            elementsfrommatrix=False,
            fluorates=mcafit._fluoRates,  # internal speed up
            addinfo=True,
        )  # additional important info

        #        print(p)
        #        import yaml
        #        with open("/tmp_14_days/chevremo/ewoksfit.yml", 'w') as fd:
        #            yaml.dump(fit, fd)

        egy = fit["energy"]
        bkg = fit["continuum"]
        grps = fit["groups"]
        yfit = fit["yfit"]
        ydata = fit["ydata"]

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            nexus_utils.now(),
            {"fit_config": config_txt, "matrix_factor": matrix_factor},
            default_levels=(output_root_uri.data_path(), output_root_group, "sumfit"),
        ) as (process_group, already_existed):

            data_grp = process_group.require_group("fitted_par")
            data_grp.attrs["NX_class"] = "NXcollection"
            for k, v in zip(fit["parameters"], fit["fittedpar"]):
                if k in data_grp:
                    del data_grp[k]
                data_grp[k] = v

            data_grp = process_group.require_group("sigma_par")
            data_grp.attrs["NX_class"] = "NXcollection"
            for k, v in zip(fit["parameters"], fit["sigmapar"]):
                if k in data_grp:
                    del data_grp[k]
                data_grp[k] = v

            data_grp = process_group.require_group("areal_density")
            data_grp.attrs["NX_class"] = "NXcollection"
            for k, v in conc["mass fraction"].items():
                if k in data_grp:
                    del data_grp[k]
                data_grp[k] = v * matrix_factor
                data_grp[k].attrs["units"] = "ng/mm2"

            data_grp = process_group.require_group("fitarea")
            data_grp.attrs["NX_class"] = "NXcollection"
            for k, v in conc["fitarea"].items():
                if k in data_grp:
                    del data_grp[k]
                data_grp[k] = v

            _res = process_group.require_group("spectrum")

            _res.attrs["NX_class"] = "NXdata"
            _res.attrs["signal"] = "fit"
            _res.attrs["auxiliary_signals"] = [
                "background",
                "spectrum",
            ] + grps
            _res.attrs["axes"] = ["energy"]

            # Remove previous result if any
            addnames = ("fit", "spectrum", "background", "energy", "pileup")
            for n in addnames:
                if n in _res:
                    del _res[n]

            egyds = _res.create_dataset("energy", data=egy)
            egyds.attrs["units"] = "keV"

            _res.create_dataset("fit", data=yfit)
            _res.create_dataset("spectrum", data=ydata)
            _res.create_dataset("background", data=bkg)

            if "pileup" in fit:
                _res.create_dataset("pileup", data=fit["pileup"])

            for g in grps:
                if g in _res:
                    del _res[g]
                _res.create_dataset(g, data=fit[f"y{g}"])

            # Set defaults
            process_group.attrs["default"] = "spectrum"
            process_group.parent.attrs["default"] = "sumfit"
            process_group.parent.parent.attrs["default"] = output_root_group

        if gallery_filename is not None:
            figpath = os.path.join(
                os.path.dirname(output_root_uri.file_path()), "gallery"
            )
            gallery_filename = os.path.join(figpath, os.path.basename(gallery_filename))
            os.makedirs(figpath, exist_ok=True)
            os.chmod(figpath, 0o770)  # nosec B103: group write required!

            fig, ax = plt.subplots(figsize=(8, 5), layout="constrained")
            colors = plt.get_cmap("nipy_spectral")(np.linspace(0.1, 0.9, len(grps)))

            ax.plot(egy, ydata, "-", color="0.7", label="Data", linewidth=4)
            # ax.plot(egy, yfit, 'r-', label="Fit", linewidth=2)

            for c, g in zip(colors, sorted(grps)):
                ax.plot(egy, fit[f"y{g}"] + bkg, "--", label=g, linewidth=1.5, color=c)

            if "pileup" in fit:
                ax.plot(
                    egy,
                    fit["pileup"] + bkg,
                    ":",
                    color="tab:blue",
                    label="Pileup",
                    linewidth=1.5,
                )

            ax.plot(egy, bkg, "b", label="Background", linewidth=2)

            fig.legend(loc="outside right upper")

            ax.set_yscale("log")
            ax.set_xlabel("Energy (keV)")

            fig.savefig(gallery_filename)

        self.outputs.output_root_uri = (
            f"{output_root_uri.file_path()}::{output_root_uri.data_path()}"
        )
        self.outputs.bliss_scan_uri = (
            f"{bliss_scan_uri.file_path()}::{bliss_scan_uri.data_path()}"
        )
        self.outputs.detector_name = detector_name
        self.outputs.config_file = config_file
        self.outputs.output_root_group = output_root_group



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 25 16:19:55 2026

@author: blissadm
"""

import os
from datetime import datetime as dt

import numpy as np
import pint
from esrf_pathlib import ESRFPath
from ewokscore import Task
from nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey
from silx.io import h5py_utils
from silx.io import url

ureg = pint.UnitRegistry()


def _tostr(val):
    if isinstance(val, str):
        return val
    elif isinstance(val, bytes):
        return val.decode()
    else:
        return str(val)


DEFAULTS = {
    "include_positioners": True,
    "scan_list": None,
    "motors_mapping": {
        "x_translation": ["isy", "spy"],
        "y_translation": ["isz", "spz"],
        "z_translation": [
            "sx",
        ],
        "rotation_angle": [
            "axis_somega",
        ],
    },
    "motors_units": {
        "isy": "nm",
        "isz": "nm",
    },
}


class FluoSinogram2Nx(
    Task,
    input_names=[
        "output_root_uri",
        "output_root_group",
    ],
    optional_input_names=[
        "include_positioners",
        "scan_list",
        "motors_mapping",
        "motors_units",
    ],
    output_names=["output_file", "output_root_uri", "output_root_group"],
):
    """
    This task takes a fulltomo sequence and creates the associated NXtomo file.
    """

    def run(self):

        pars = {**DEFAULTS, **self.get_input_values()}

        output_root_uri = url.DataUrl(pars["output_root_uri"])
        output_root_group = pars["output_root_group"]
        motors_mapping = pars["motors_mapping"]
        include_positioners = pars["include_positioners"]
        motors_units = pars["motors_units"]

        esrf_path = ESRFPath(output_root_uri.file_path())
        scan_no = int(output_root_uri.data_path().split("/")[1].split(".")[0])

        output_path = os.path.join(esrf_path.processed_dataset_path, "projections")
        output_file = os.path.join(
            output_path, f"{esrf_path.collection}_{esrf_path.dataset}_{scan_no:04d}.nx"
        )

        if not os.path.isdir(output_path):
            os.makedirs(output_path, exist_ok=True)
            os.chmod(
                output_path, 0o770
            )  # nosec B103: group write required for shared pipeline

        egy = None

        print(output_root_uri)

        with h5py_utils.open_item(
            output_root_uri.file_path(), output_root_uri.data_path()
        ) as scan:

            if (
                "metadata" in scan["instrument"]
                and "InstrumentMonochromator_energy" in scan["instrument/metadata"]
            ):
                egy = ureg.Quantity(
                    float(
                        _tostr(
                            scan["instrument/metadata/InstrumentMonochromator_energy"][
                                ()
                            ]
                        )
                    ),
                    "keV",
                )
            elif "InstrumentMonochromator" in scan["instrument"]:
                egy = ureg.Quantity(
                    float(
                        _tostr(scan["instrument/InstrumentMonochromator/energy"][()])
                    ),
                    "keV",
                )

            positioners = {}
            positioners_units = {}
            rays = set()

            count_time = float(
                scan["instrument/fscan_parameters/acq_time"][()].decode()
            )
            fast_pixel_size = float(
                scan["instrument/fscan_parameters/fast_step_size"][()].decode()
            )
            slow_n = int(scan["instrument/fscan_parameters/slow_npoints"][()].decode())
            fast_n = int(scan["instrument/fscan_parameters/slow_npoints"][()].decode())
            fast_motor_name = scan["instrument/fscan_parameters/fast_motor"][
                ()
            ].decode()
            slow_motor_name = scan["instrument/fscan_parameters/slow_motor"][
                ()
            ].decode()

            for k in scan["instrument/positioners"]:
                ds = scan["instrument/positioners"][k]

                positioners[k] = np.empty((slow_n,), dtype=np.float64) * np.nan
                positioners_units[k] = motors_units.get(k, ds.attrs.get("units", None))

                vals = ds[()]

                if isinstance(vals, np.ndarray) and len(vals) == slow_n * fast_n:
                    vals = vals[::fast_n]

                positioners[k][:] = vals

            if slow_motor_name != "somega":
                raise RuntimeError("Expected rotation as slow motor!")

            fast_pixel_size = ureg.Quantity(
                fast_pixel_size, positioners_units[fast_motor_name]
            )

            # Start populating the NXtomo object
            nxtomo = NXtomo()
            nxtomo.start_time = dt.now()
            nxtomo.bliss_original_files = (pars["output_root_uri"],)

            nxtomo.energy = egy
            # TODO
            nxtomo.instrument.detector.tomo_n = slow_n
            #            nxtomo.instrument.detector.distance =
            nxtomo.instrument.detector.x_pixel_size = fast_pixel_size
            nxtomo.instrument.detector.y_pixel_size = fast_pixel_size
            nxtomo.sample.x_pixel_size = fast_pixel_size
            nxtomo.sample.y_pixel_size = fast_pixel_size
            #            nxtomo.instrument.source.distance =
            #            nxtomo.sample.propagation_distance =

            nxtomo.sample.name = esrf_path.collection
            nxtomo.title = esrf_path.dataset

            #            nxtomo.control.data =

            nxtomo.instrument.name = esrf_path.beamline
            nxtomo.instrument.detector.count_time = pint.Quantity(
                np.array(count_time), "s"
            )
            nxtomo.instrument.detector.image_key_control = [
                ImageKey.PROJECTION,
            ] * slow_n

            for k, mots in motors_mapping.items():

                values = pint.Quantity(
                    np.zeros((slow_n,), dtype=np.float64),
                    "degree" if k == "rotation_angle" else "mm",
                )

                for m in mots:
                    #                    print(m, positioners_units[m])
                    values += pint.Quantity(positioners[m], positioners_units[m])

                setattr(nxtomo.sample, k, values)

            regrid = scan[f"{output_root_group}/regrid/results/massfractions"]

            rays = set()
            rays.add(regrid.attrs["signal"])

            for sig in regrid.attrs["auxiliary_signals"]:
                rays.add(sig)

            for elm in rays:

                data = regrid[elm][()]

                data[np.isnan(data)] = 0

                data.shape = data.shape[0], 1, data.shape[1]

                nxtomo.instrument.detector.data = (
                    data * 1e-7
                )  # Conversion from ng/mm2 to g/cm2
                nxtomo.end_time = dt.now()
                nxtomo.save(output_file, f"{output_root_group}_{elm}", overwrite=True)

        if include_positioners:
            for elm in rays:
                with h5py_utils.open_item(
                    output_file, f"/{output_root_group}_{elm}", mode="a"
                ) as fd:

                    grp = fd.require_group("instrument/positioners")
                    grp.attrs["NX_class"] = "NXcollection"

                    for k, v in positioners.items():
                        if k in grp:
                            del grp[k]

                        grp[k] = v
                        if positioners_units[k] is not None:
                            grp[k].attrs["units"] = positioners_units[k]

        self.outputs.output_root_uri = pars["output_root_uri"]
        self.outputs.output_root_group = pars["output_root_group"]
        self.outputs.output_file = output_file



import os
from datetime import datetime as dt

import numpy as np
import pint
from esrf_pathlib import ESRFPath
from ewokscore import Task
from h5py import VirtualSource
from nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey
from silx.io import h5py_utils
from silx.io.url import DataUrl

ureg = pint.UnitRegistry()


def _tostr(val):
    if isinstance(val, str):
        return val
    elif isinstance(val, bytes):
        return val.decode()
    else:
        return str(val)


class Tomo2Nx(
    Task,
    input_names=[
        "bliss_scan_uri",
    ],
    optional_input_names=[
        "output_id",
        "include_positioners",
        "position_mapping_override",
        "scan_list",
    ],
    output_names=["bliss_scan_uri", "output_root_uri"],
):
    """
    This task takes a fulltomo sequence and creates the associated NXtomo file.
    """

    def run(self):
        bliss_uri = self.inputs["bliss_scan_uri"]
        uri = bliss_uri.split("::/")

        bliss_filename = uri[0]
        scan_id = uri[1]

        scan_no = int(scan_id.split(".")[0])

        esrf_path = ESRFPath(bliss_filename)

        scan_list = self.get_input_value("scan_list", None)

        output_id = self.get_input_value("output_id", f"{scan_no:04d}")

        output_file = os.path.join(
            esrf_path.processed_dataset_path,
            "projections",
            f"{esrf_path.collection}_{esrf_path.dataset}_{output_id}.nx",
        )

        output_entry = f"entry_{output_id}"

        output_path = os.path.abspath(os.path.dirname(output_file))

        if not os.path.isdir(output_path):
            os.makedirs(output_path, exist_ok=True)
            os.chmod(
                output_path, 0o770
            )  # nosec B103: group write required for shared pipeline

        bliss_positioners_mapping = {
            "sample_u": "sample_u",
            "sample_v": "sample_v",
            "z_translation": "translation_x",
            "x_translation": "translation_y",
            "y_translation": "translation_z",
            "rotation_angle": "rotation",
        }
        motors_mapping = self.get_input_value("position_mapping_override", {})

        types_mapping = {
            "tomo:dark": ImageKey.DARK_FIELD,
            "tomo:flat": ImageKey.FLAT_FIELD,
            "tomo:step": ImageKey.PROJECTION,
            "tomo:return_ref": ImageKey.ALIGNMENT,
        }

        nxtomo = NXtomo()
        nxtomo.start_time = dt.now()
        nxtomo.bliss_original_files = (bliss_uri,)
        sy_pos = None
        dety = None

        positioners_start = {}

        # Wait for scan completion
        with h5py_utils.open_item(
            bliss_filename, f"{scan_id}/end_time", retry_invalid=True
        ):
            pass

        # open the tomo seauence, read the configuration and scan sequence
        with h5py_utils.open_item(bliss_filename, scan_id) as tomo_sequence:
            nxtomo.energy = pint.Quantity(
                tomo_sequence["technique/scan/energy"][()], "keV"
            )
            nxtomo.instrument.detector.tomo_n = tomo_sequence["technique/scan/tomo_n"][
                ()
            ]

            sdist = tomo_sequence["technique/scan/sample_detector_distance"]
            nxtomo.instrument.detector.distance = ureg.Quantity(
                sdist[()], sdist.attrs.get("units", "mm")
            )

            ops_ds = tomo_sequence["technique/optic/optics_pixel_size"]
            ops = ureg.Quantity(ops_ds[()], ops_ds.attrs.get("units", "um"))
            nxtomo.instrument.detector.x_pixel_size = ops
            nxtomo.instrument.detector.y_pixel_size = ops

            sps_ds = tomo_sequence["technique/optic/sample_pixel_size"]
            sps = ureg.Quantity(sps_ds[()], ops_ds.attrs.get("units", "um"))
            nxtomo.sample.x_pixel_size = sps
            nxtomo.sample.y_pixel_size = sps

            ssd_ds = tomo_sequence["technique/scan/source_sample_distance"]
            ssd = ureg.Quantity(-ssd_ds[()], ssd_ds.attrs.get("units", "mm"))
            nxtomo.instrument.source.distance = ssd

            effd_ds = tomo_sequence["technique/scan/effective_propagation_distance"]
            effd = ureg.Quantity(effd_ds[()], effd_ds.attrs.get("units", "mm"))
            nxtomo.sample.propagation_distance = effd

            scan_numbers = tomo_sequence["subscans"]["scan_numbers"][:]
            detector = tomo_sequence["technique/tomoconfig/detector"][0]

            for k in tomo_sequence["instrument/positioners"]:
                ds = tomo_sequence["instrument/positioners"][k]
                positioners_start[k] = ureg.Quantity(
                    ds[()], ds.attrs.get("units", None)
                )

            for k, m in bliss_positioners_mapping.items():

                if k in motors_mapping:
                    continue

                if m in tomo_sequence["technique/tomoconfig"]:
                    motors_mapping[k] = []

                    tot = None

                    for v in tomo_sequence["technique/tomoconfig"][m]:
                        pos_ds = tomo_sequence["instrument/positioners"][v]
                        units = pos_ds.attrs.get("units", None)
                        motors_mapping[k] += [(_tostr(v), units)]

                        if tot is None:
                            tot = ureg.Quantity(pos_ds[()], units)
                        else:
                            tot += ureg.Quantity(pos_ds[()], units)

                    if m == "translation_y":
                        sy_pos = tot

            detector_y_ds = tomo_sequence["technique/tomoconfig/detector_center_y"]
            dety = ureg.Quantity(
                detector_y_ds[()], detector_y_ds.attrs.get("units", "mm")
            )

            types = []

            if (
                "technique" in tomo_sequence
                and "subscans" in tomo_sequence["technique"]
            ):

                keys = [k for k in tomo_sequence["technique"]["subscans"]]
                keys.sort(key=lambda v: int(v[4:]))

                for k in keys:
                    typ = tomo_sequence["technique"]["subscans"][k]["type"][()]
                    #                    print(k, typ)
                    if isinstance(typ, bytes):
                        typ = typ.decode()
                    types += [
                        typ,
                    ]
            else:  # Fallback on titles
                for i, n in enumerate(scan_numbers):
                    title = tomo_sequence.parent[f"{n}.1"]["title"][()].decode()

                    if "dark" in title:
                        types += [
                            "tomo:dark",
                        ]
                    elif "flat" in title:
                        types += [
                            "tomo:flat",
                        ]
                    elif "static" in title:
                        types += [
                            "tomo:return_ref",
                        ]
                    else:
                        types += [
                            "tomo:step",
                        ]

        detector_data_url = []
        image_key_control = []
        positioners = {k: [] for k in motors_mapping}
        srcur = []
        count_time = []

        #        print(motors_mapping)

        # Go so subscans and get infos
        #        print(scan_numbers)

        if scan_list is not None:
            _scan_numbers = scan_numbers.copy()
            _types = types.copy()

            scan_numbers = []
            types = []

            for i, s in enumerate(_scan_numbers):
                if s in scan_list:
                    scan_numbers += [
                        s,
                    ]
                    types += [
                        _types[i],
                    ]

        shape = None

        for i, n in enumerate(scan_numbers):
            if types[i] not in types_mapping:  # Maybe a custom scan in the sequence
                continue

            with h5py_utils.open_item(bliss_filename, f"/{n}.1") as scan:
                det_ds = scan["measurement"][detector]
                shape = det_ds.shape
                n_images = det_ds.shape[0]

                pos = {k: np.zeros((n_images,), dtype=np.float64) for k in positioners}

                for k, v in motors_mapping.items():
                    for m in v:
                        posval = ureg.Quantity(
                            scan["instrument/positioners"][m[0]][()], m[1]
                        )
                        pos[k] += posval

                    positioners[k] += [
                        pos[k],
                    ]

                if "current" in scan["measurement"]:
                    srcur += [scan["measurement/current"][()]]

                count_time_ds = scan["scan_parameters/count_time"]
                count_time += [
                    ureg.Quantity(
                        np.array(
                            [
                                count_time_ds[()],
                            ]
                            * n_images
                        ),
                        count_time_ds.attrs.get("units", "s"),
                    ),
                ]

                image_key_control += [
                    types_mapping[types[i]],
                ] * n_images
                if (
                    det_ds.is_virtual
                ):  # Resolve virtual datasets to avoid any pointers to bliss scan file if possible
                    for vs in det_ds.virtual_sources():
                        src_file = vs.file_name
                        src_path = vs.dset_name

                        if not os.path.isabs(src_file):
                            src_file = os.path.join(
                                os.path.dirname(bliss_filename), src_file
                            )

                        # src_file = os.path.relpath(src_file, output_path)

                        # print(src_file)

                        detector_data_url += [
                            VirtualSource(src_file, src_path, vs.src_space.shape),
                        ]
                else:
                    detector_data_url += [
                        DataUrl(
                            file_path=bliss_filename,
                            data_path=f"/{n}.1/measurement/{detector}",
                            scheme="silx",
                        ),
                    ]

        if sy_pos is not None and dety is not None:
            rotaxis_pos = ((sy_pos - dety) / sps).to_reduced_units()
            nxtomo.instrument.detector.x_rotation_axis_pixel_position = (
                shape[1] // 2 - rotaxis_pos.magnitude
            )

        # Sample positioners
        extra_positioners = []

        for k, v in positioners.items():
            try:
                setattr(nxtomo.sample, k, np.concatenate(v))
            except Exception as e:
                print(e)
                extra_positioners += [
                    k,
                ]

        nxtomo.sample.name = esrf_path.collection
        nxtomo.title = esrf_path.dataset

        nxtomo.control.data = np.concatenate(srcur)

        nxtomo.instrument.name = esrf_path.beamline
        nxtomo.instrument.detector.data = detector_data_url
        nxtomo.instrument.detector.count_time = np.concatenate(count_time)
        nxtomo.instrument.detector.image_key_control = image_key_control

        nxtomo.end_time = dt.now()
        nxtomo.save(output_file, output_entry, overwrite=True)

        if self.get_input_value("include_positioners", True):
            with h5py_utils.open_item(output_file, output_entry, mode="a") as fd:

                grp = fd.require_group("instrument/positioners")
                grp.attrs["NX_class"] = "NXcollection"

                for k, v in positioners_start.items():
                    if k in grp:
                        del grp[k]

                    grp[k] = np.float32(v.magnitude)
                    if v.units is not None:
                        grp[k].attrs["units"] = f"{v.units:~P}"

        self.outputs["bliss_scan_uri"] = self.inputs["bliss_scan_uri"]
        self.outputs["output_root_uri"] = output_file + f"::/{output_entry}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 10 21:46:16 2026

@author: blissadm
"""

# import time

import numpy as np
from ewokscore import Task

# from ewoksfluo.io.hdf5 import split_h5uri
from ewoksfluo.tasks import nexus_utils
from ewoksfluo.tasks.hdf5_utils import create_hdf5_link
from PyMca5.PyMcaIO import ConfigDict
from silx.io import h5py_utils
from silx.io import url

DEFAULTS = {
    "stripped_detector_name": None,
    "xrf_spectra_uri_template": "instrument/{}/data",
    "photon_counter_name": None,
}


class StripBackgroundFromSumfit(
    Task,
    input_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "detector_name",
        "config_file",
    ],
    optional_input_names=[
        "stripped_detector_name",
        "xrf_spectra_uri_template",
        "photon_counter_name",
    ],
    output_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "detector_name",
        "config_file",
    ],
):
    def run(self):

        params = {**DEFAULTS, **self.get_input_values()}

        bliss_scan_uri = url.DataUrl(params["bliss_scan_uri"])
        output_root_uri = url.DataUrl(params["output_root_uri"])
        detector_name = params["detector_name"]
        config_file = params["config_file"]
        stripped_detector_name = params["stripped_detector_name"]
        photon_counter_name = params["photon_counter_name"]

        if stripped_detector_name is None:
            stripped_detector_name = f"{detector_name}_stripped"

        xrf_spectra_uri_template = params["xrf_spectra_uri_template"]

        # Read min and max from config
        cfg = ConfigDict.ConfigDict(filelist=config_file)
        xmin = cfg["fit"]["xmin"]
        xmax = cfg["fit"]["xmax"]

        photon_counter = None

        with h5py_utils.open_item(bliss_scan_uri.file_path(), bliss_scan_uri.data_path()) as scan:  # type: ignore[reportGeneralTypeIssues]

            data = np.array(scan[xrf_spectra_uri_template.format(detector_name)])
            bkg = np.array(scan["sumfit/{}/spectrum/background".format(detector_name)])

            if photon_counter_name is not None:
                photon_counter = np.array(
                    scan[xrf_spectra_uri_template.format(photon_counter_name)]
                )

        # Generate background as long as data py expending head and tail values
        lbkg = np.zeros((1, data.shape[1]), dtype=np.float32)
        lbkg[:, xmin : xmax + 1] = bkg
        lbkg[:, :xmin] = bkg[0]
        lbkg[:, xmax + 1 :] = bkg[-1]

        if photon_counter is not None:
            # Read photon count from config and scale the background accordingly
            flux = float(cfg["concentrations"]["flux"])
            time = float(cfg["concentrations"]["time"])

            cfg_photons = flux * time
            photon_factor = photon_counter / cfg_photons

            photon_factor.shape = photon_factor.shape[0], 1

            lbkg = lbkg * photon_factor

            print(lbkg.shape, data.shape)

        newdata = (
            data - lbkg
        )  # maybe needed to weight according to the flux at each pixel.

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            nexus_utils.now(),
            {},
            default_levels=(
                output_root_uri.data_path(),
                "stripped_spectra",
                stripped_detector_name,
            ),
        ) as (process_group, already_existed):
            outentry = process_group.parent.parent

            if "mcastripped" in process_group:
                del process_group["mcastripped"]

            nxdata = nexus_utils.create_nxdata(
                process_group, "mcastripped", signal="data"
            )
            nxdata.attrs["interpretation"] = "spectrum"

            if "data" in nxdata:
                del nxdata["data"]

            dset = nxdata.create_dataset(
                "data",
                data=newdata,
                chunks=(1, newdata.shape[1]),
                compression="gzip",
                shuffle=True,
            )

            if not already_existed:

                nxdetector = outentry["instrument"].create_group(stripped_detector_name)
                nxdetector.attrs["NX_class"] = "NXdetector"
                create_hdf5_link(nxdetector, "data", dset)
                create_hdf5_link(outentry["measurement"], stripped_detector_name, dset)

        self.outputs.output_root_uri = (
            f"{output_root_uri.file_path()}::{output_root_uri.data_path()}"
        )
        self.outputs.bliss_scan_uri = (
            f"{bliss_scan_uri.file_path()}::{bliss_scan_uri.data_path()}"
        )
        self.outputs.detector_name = stripped_detector_name
        self.outputs.config_file = config_file



# import h5py
# import hdf5plugin  # noqa: F401
import logging

import numpy as np
from ewokscore import Task
from ewoksfluo.io.hdf5 import ReadHdf5File
from ewoksfluo.tasks import nexus_utils
from silx.io import h5py_utils
from silx.io import url

logger = logging.getLogger(__name__)

DEFAULTS = {
    "fscan_parameters": "instrument/fscan_parameters",
    "instrument_data_template": "instrument/{}/data",
    "vertical_motor_names": ["spz", "spzp"],
}


class SpectralRegrid(
    Task,
    input_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "output_root_group",
        "detector_name",
    ],
    optional_input_names=[
        "fscan_parameters",
        "instrument_data_template",
        "vertical_motor_names",
    ],
    output_names=[
        "bliss_scan_uri",
        "output_root_uri",
        "output_root_group",
        "detector_name",
    ],
):
    """Regrid raw spectrums"""

    def run(self):
        start_time = nexus_utils.now()
        params = {
            **DEFAULTS,
            **self.get_input_values(),
        }

        bliss_scan_uri = url.DataUrl(params["bliss_scan_uri"])
        output_root_uri = url.DataUrl(params["output_root_uri"])
        output_root_group = params["output_root_group"]
        detector_name = params["detector_name"]
        fscan_parameters = params["fscan_parameters"]
        instrument_data_template = params["instrument_data_template"]
        vertical_motor_names = params["vertical_motor_names"]

        logger.warning(f"{detector_name=} {bliss_scan_uri=}")

        with h5py_utils.open_item(bliss_scan_uri.file_path(), bliss_scan_uri.data_path()) as scan:  # type: ignore[reportGeneralTypeIssues]

            fp = scan[fscan_parameters]

            fastmot = fp["fast_motor"][()].decode()
            slowmot = fp["slow_motor"][()].decode()
            mode = fp["fast_motor_mode"][()].decode()
            fast_n = int(fp["fast_npoints"][()].decode())
            slow_n = int(fp["slow_npoints"][()].decode())

            cnt = scan[instrument_data_template.format(detector_name)]
            cnt_shape = cnt.shape
            cnt_dtype = cnt.dtype

        A = np.arange(fast_n * slow_n)
        #        A.shape = (fast_n, slow_n)
        A.shape = (slow_n, fast_n)

        is_vertical_scan = False

        if fastmot in vertical_motor_names:
            A = A.swapaxes(0, 1)
            is_vertical_scan = True

        if mode == "ZIGZAG":
            A[1::2, :] = A[1::2, :][:, ::-1]

        logger.warning(f"{detector_name=} {output_root_uri=}")

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config={
                "is_vertical_scan": is_vertical_scan,
                "mode": mode,
                "fast_motor": fastmot,
                "slow_motor": slowmot,
            },
            default_levels=(
                output_root_uri.data_path(),
                output_root_group,
                "spectral_regrid",
            ),
        ) as (process_group, already_existed):

            if not already_existed:

                nxdata = nexus_utils.create_nxdata(
                    process_group, "gridspectrum", signal="data"
                )
                nxdata.attrs["interpretation"] = "image"

                dset = nxdata.create_dataset(
                    "data",
                    shape=(cnt_shape[1], *A.shape),
                    dtype=cnt_dtype,
                    chunks=(1, *A.shape),
                    shuffle=True,
                    compression="gzip",
                )

                with ReadHdf5File(
                    bliss_scan_uri.file_path()
                ) as fd:  # Maybe the file is already opened in append mode by the save_in_ewoks_process.
                    scan = fd[bliss_scan_uri.data_path()]

                    src = scan[instrument_data_template.format(detector_name)][()]

                    if src.shape[0] < fast_n * slow_n:  # Fix in case of user abortion
                        _src = src
                        src = np.zeros((fast_n * slow_n, src.shape[1]), src.dtype)
                        src[: _src.shape[0]] = _src

                    #                    ds = np.empty(shape=(cnt_shape[1], *A.shape), dtype=cnt_dtype)
                    #                    for i in range(cnt_shape[0]):
                    #                        ds[(slice(None), *np.unravel_index(i, A.shape))] = src[A.flat[i]]

                    # Vectorized version
                    ds = np.moveaxis(src[A], -1, 0)

                dset[:] = ds

        self.outputs.bliss_scan_uri = bliss_scan_uri
        self.outputs.output_root_uri = output_root_uri
        self.outputs.output_root_group = output_root_group
        self.outputs.detector_name = detector_name
